from __future__ import annotations

from typing import TYPE_CHECKING, Iterator, Protocol

from .cards import Card, Suit

if TYPE_CHECKING:
    from .bet import Bet
    from .round_simulator import _Player


class WhistEventListener(Protocol):
    def notify_round_start(
        self,
        revealed_card: Card,
        trump_suit: Suit,
        hand_size: int,
        players: list[_Player],
    ) -> None:
        """Called when a new round starts."""
        return None

    def notify_card_played(self, player: _Player, played_card: Card) -> None:
        """Called when a player plays any card."""
        return None

    def notify_trick_won(self, player: _Player) -> None:
        """Called when a trick is won."""
        return None

    def notify_bets_made(self, bets: Iterator[tuple[_Player, Bet]]) -> None:
        """Called when all bets have been finalised."""
        return None

    def notify_new_trick(self, player_id: int) -> None:
        """Called when a trick is about to start (starting with `player_id`)."""
        return None



import pytest

from whist_backend.cards import Card, Rank, Suit, new_deck, new_shuffled_deck


@pytest.mark.parametrize(
    ["rank1", "rank2", "expected"],
    [
        (Rank.ACE, Rank.KING, True),
        (Rank.KING, Rank.QUEEN, True),
        (Rank.QUEEN, Rank.JACK, True),
        (Rank.JACK, Rank.TEN, True),
        (Rank.TEN, Rank.NINE, True),
        (Rank.NINE, Rank.TWO, True),
        (Rank.ACE, Rank.ACE, False),
        (Rank.KING, Rank.KING, False),
        (Rank.QUEEN, Rank.QUEEN, False),
        (Rank.TEN, Rank.TEN, False),
        (Rank.TWO, Rank.TWO, False),
        (Rank.TWO, Rank.THREE, False),
        (Rank.TEN, Rank.JACK, False),
        (Rank.JACK, Rank.QUEEN, False),
        (Rank.QUEEN, Rank.KING, False),
        (Rank.KING, Rank.ACE, False),
        (Rank.TWO, Rank.ACE, False),
    ],
)
def test_rank_inequality(rank1: Rank, rank2: Rank, expected: bool) -> None:
    assert (rank1 > rank2) is expected


@pytest.mark.parametrize(
    ["rank", "expected"],
    [
        (Rank.ACE, "A"),
        (Rank.KING, "K"),
        (Rank.QUEEN, "Q"),
        (Rank.JACK, "J"),
        (Rank.TEN, "10"),
        (Rank.THREE, "3"),
        (Rank.TWO, "2"),
    ],
)
def test_rank_string(rank: Rank, expected: str) -> None:
    assert str(rank) == expected


@pytest.mark.parametrize(
    ["suit", "expected"],
    [
        (Suit.SPADES, "♠"),
        (Suit.HEARTS, "♥"),
        (Suit.DIAMONDS, "♦"),
        (Suit.CLUBS, "♣"),
    ],
)
def test_suit_string(suit: Suit, expected: str) -> None:
    assert str(suit) == expected


@pytest.mark.parametrize(
    ["card", "expected"],
    [
        (Card(Rank.ACE, Suit.SPADES), " A♠"),
        (Card(Rank.KING, Suit.CLUBS), " K♣"),
        (Card(Rank.QUEEN, Suit.DIAMONDS), " Q♦"),
        (Card(Rank.JACK, Suit.SPADES), " J♠"),
        (Card(Rank.TEN, Suit.HEARTS), "10♥"),
        (Card(Rank.SIX, Suit.SPADES), " 6♠"),
    ],
)
def test_card_string(card: Card, expected: str) -> None:
    assert str(card) == expected


def test_deck() -> None:
    assert len(set(new_deck())) == 52
    assert len(set(new_deck() + new_deck())) == 52


def test_shuffled_deck() -> None:
    assert set(new_deck()) == set(new_shuffled_deck())



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from itertools import product
from random import shuffle


class Suit(Enum):
    """An enum for each card suit.

    .. automethod:: __str__
    """

    SPADES = 0
    HEARTS = 1
    DIAMONDS = 2
    CLUBS = 3

    def __str__(self) -> str:
        """Return a single character representing the suit, eg `♠`."""
        return "♠♥♦♣"[self.value]


class Rank(Enum):
    """An Enum for each card rank.

    Note that ACE has value 1.

    You can use the `>` operator on Ranks.

    .. automethod:: __str__

    .. automethod:: __lt__

    .. automethod:: __gt__

    .. automethod:: __le__

    .. automethod:: __ge__
    """

    value: int

    ACE = 1
    TWO = 2
    THREE = 3
    FOUR = 4
    FIVE = 5
    SIX = 6
    SEVEN = 7
    EIGHT = 8
    NINE = 9
    TEN = 10
    JACK = 11
    QUEEN = 12
    KING = 13

    def __lt__(self, other: Rank) -> bool:
        """Given `self < other`, return `True` if `other` is strictly better."""
        if not isinstance(other, Rank):
            NotImplementedError("Ranks can only be compared to ranks.")  # type: ignore

        return self.whist_value() < other.whist_value()

    def __str__(self) -> str:
        """Return a short string representation of the rank, eg `A` or `10`."""
        if self == Rank.ACE:
            return "A"
        elif self == Rank.JACK:
            return "J"
        elif self == Rank.QUEEN:
            return "Q"
        elif self == Rank.KING:
            return "K"
        else:
            return str(self.value)

    def whist_value(self) -> int:
        return 14 if self == Rank.ACE else self.value


@dataclass(frozen=True)
class Card:
    """Represents a card object.

    This is implemented as a frozen dataclass, so magic methods such as `__hash__` and
    `__eq__` are automatically defined on it.

    .. automethod:: __str__
    """

    rank: Rank
    suit: Suit

    def __str__(self) -> str:
        """Return a 3 character representation of the card, eg ` A♠`."""
        return f"{self.rank:>2}{self.suit}"


def new_deck() -> list[Card]:
    """Return a full, new, sorted deck of cards."""
    return [Card(rank, suit) for rank, suit in product(Rank, Suit)]


def new_shuffled_deck() -> list[Card]:
    """Return a full, new, shuffled deck of cards."""
    deck = new_deck()
    shuffle(deck)
    return deck



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Sequence, TypeVar

from .cards import Card, Suit, new_shuffled_deck
from .event_listener_protocol import WhistEventListener

if TYPE_CHECKING:
    from .bet import Bet
    from .player_strategy_protocol import PlayerStrategy


T = TypeVar("T")


class _Player:
    """This is used to ensure that the rules of whist are obeyed by the
    :class:`whist_backend.player_strategy_protocol.PlayerStrategy`s.
    """

    def __init__(
        self,
        strategy: PlayerStrategy,
        hand: list[Card],
        revealed_card: Card,
        player_id: int,
        number_of_players: int,
    ) -> None:
        self.strategy = strategy
        self.strategy.notify_round_start(
            hand, revealed_card.suit, player_id, revealed_card, number_of_players
        )
        self.hand = hand
        self.played_cards: list[Card] = []
        self.player_id = player_id
        self.tricks_won = 0

    def make_move(self, initial_suit: Optional[Suit]) -> Card:
        allowed_cards = self.get_allowed_cards(initial_suit)
        picked_card = self.strategy.make_move(allowed_cards)
        assert picked_card in allowed_cards, (
            f"{self.strategy.name} picked an invalid card (picked {picked_card} which"
            f"is not one of {allowed_cards})."
        )
        self.played_cards.append(picked_card)
        self.hand.remove(picked_card)
        return picked_card

    def get_allowed_cards(self, initial_suit: Optional[Suit]) -> list[Card]:
        if cards_of_initial_suit := [
            card for card in self.hand if card.suit == initial_suit
        ]:
            return cards_of_initial_suit
        else:
            return self.hand

    def make_bet(self, disallowed_bet: Optional[int] = None) -> Bet:
        bet = self.strategy.make_bet(disallowed_bet)
        assert (
            bet.number_of_tricks != disallowed_bet
        ), f"{self.strategy.name} made an invalid bet."
        return bet

    def notify_trick_won(self, player_id: int) -> None:
        if player_id == self.player_id:
            self.tricks_won += 1
        self.strategy.notify_trick_won(player_id)

    def __str__(self) -> str:
        cards_left = ",".join(str(card) for card in self.hand)
        cards_played = ",".join(str(card) for card in self.played_cards)
        return f"{self.strategy.name}: hand {cards_left}" + (
            f" (played {cards_played})" if cards_played else ""
        )


def deal_hands(
    number_of_players: int, hand_size: int, deck: list[Card]
) -> list[list[Card]]:
    return [[deck.pop() for _ in range(hand_size)] for _ in range(number_of_players)]


def determine_trump_suit_and_hand_sizes(card: Card) -> tuple[Suit, int]:
    trump_suit = card.suit
    card_value = card.rank.value
    if card_value < 5:
        return trump_suit, 10 - card_value
    elif card_value > 10:
        return trump_suit, 21 - card_value
    else:
        return trump_suit, card_value


def cyclic_shift(items: list[T], shift_amount: int) -> list[T]:
    shift_amount %= len(items)
    return items[shift_amount:] + items[:shift_amount]


def find_trick_winner(
    cards_and_players: list[tuple[_Player, Card]], trump_suit: Suit, initial_suit: Suit
) -> _Player:
    trump_suit_moves = [
        move for move in cards_and_players if move[1].suit == trump_suit
    ]
    initial_suit_moves = [
        move for move in cards_and_players if move[1].suit == initial_suit
    ]
    if trump_suit_moves:
        return max(trump_suit_moves, key=lambda move: move[1].rank)[0]
    else:
        return max(initial_suit_moves, key=lambda move: move[1].rank)[0]


class RoundSimulator:
    def __init__(
        self,
        strategies: Sequence[PlayerStrategy],
        *,
        event_listeners: Optional[list[WhistEventListener]] = None,
    ) -> None:
        deck = new_shuffled_deck()
        revealed_card = deck.pop()
        self.trump_suit, self.hand_size = determine_trump_suit_and_hand_sizes(
            revealed_card
        )
        number_of_players = len(strategies)
        hands = deal_hands(number_of_players, self.hand_size, deck)
        self.players = [
            _Player(strategy, hand, revealed_card, player_id, number_of_players)
            for player_id, (strategy, hand) in enumerate(zip(strategies, hands))
        ]
        self.next_starting_player = 0
        self.event_listeners = event_listeners or []
        for event_listener in self.event_listeners:
            event_listener.notify_round_start(
                revealed_card, self.trump_suit, self.hand_size, self.players
            )

    def play_round(self) -> list[tuple[PlayerStrategy, int]]:
        bets = self.gather_bets_for_round(self.players)
        for _ in range(self.hand_size):
            self.play_trick()
        results = [
            (player.strategy, bet.compute_score(player.tricks_won))
            for player, bet in zip(self.players, bets)
        ]
        for player in self.players:
            player.strategy.notify_round_end(
                [(i, score) for i, (_, score) in enumerate(results)]
            )
        return results

    def make_move_on_player(
        self, player: _Player, initial_suit: Optional[Suit]
    ) -> Card:
        picked_card = player.make_move(initial_suit)
        for other_player in self.players:
            other_player.strategy.notify_new_move(player.player_id, picked_card)
        for event_listener in self.event_listeners:
            event_listener.notify_card_played(player, picked_card)
        return picked_card

    def on_trick_win(self, player: _Player) -> None:
        for other_player in self.players:
            other_player.notify_trick_won(player.player_id)
        self.next_starting_player = player.player_id
        for event_listener in self.event_listeners:
            event_listener.notify_trick_won(player)

    def ask_player_for_bet(
        self, player: _Player, disallowed_bet: Optional[int] = None
    ) -> Bet:
        bet = player.make_bet(disallowed_bet)
        for other_player in self.players:
            other_player.strategy.notify_new_bet(player.player_id, bet)
        return bet

    def gather_bets_for_round(self, player_order: list[_Player]) -> list[Bet]:
        bets = [self.ask_player_for_bet(player) for player in player_order[:-1]]
        disallowed_bet_for_last_player = self.hand_size - sum(
            map(lambda bet: bet.number_of_tricks, bets)
        )
        bets.append(
            self.ask_player_for_bet(player_order[-1], disallowed_bet_for_last_player)
        )
        for event_listener in self.event_listeners:
            event_listener.notify_bets_made(zip(player_order, bets))
        return bets

    def play_trick(self) -> None:
        for player in self.players:
            player.strategy.notify_new_trick(self.next_starting_player)
        for event_listener in self.event_listeners:
            event_listener.notify_new_trick(self.next_starting_player)
        player_order = cyclic_shift(self.players, self.next_starting_player)
        cards_played = [self.make_move_on_player(player_order[0], None)]
        initial_suit = cards_played[0].suit
        cards_played.extend(
            self.make_move_on_player(player, initial_suit)
            for player in player_order[1:]
        )
        trick_winner = find_trick_winner(
            list(zip(player_order, cards_played)), self.trump_suit, initial_suit
        )
        self.on_trick_win(trick_winner)



from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, Optional, Protocol

if TYPE_CHECKING:
    from .bet import Bet
    from .cards import Card, Suit


class PlayerStrategy(Protocol):
    """Dictates a strategy for a bot.

    If you inherit from this class, then the only methods that have to be implemented
    are `make_move` and `make_bet`. In these methods, you must obey the rules of whist
    (this will be checked). All other methods don't expect any return value, and are
    only there so that your strategy can monitor the progress of the game.

    See the docstrings of each method for more information.
    """

    name: str

    def notify_round_start(
        self,
        hand: list[Card],
        trump_suit: Suit,
        player_id: int,
        revealed_card: Card,
        number_of_players: int,
    ) -> None:
        """Notifies when a round is about to start (i.e. cards have been dealt)."""
        return

    @abstractmethod
    def make_move(self, allowed_cards: list[Card]) -> Card:
        """Return the Card to play on this move. Must be one of `allowed_cards`."""

    def notify_new_move(self, player_id: int, card: Card) -> None:
        """Notify that `player_id` just played `card`."""
        return

    @abstractmethod
    def make_bet(self, disallowed_bet: Optional[int] = None) -> Bet:
        """Return a Bet, with number_of_tricks different to `disallowed_bet`."""

    def notify_new_bet(self, player_id: int, bet: Bet) -> None:
        """Notify that `player_id` made a bet `bet`."""
        return

    def notify_trick_won(self, player_id: int) -> None:
        """Notify that a trick has just been won by `player_id`."""
        return

    def notify_new_trick(self, player_id: int) -> None:
        """Notify that a new trick is about to start, starting with `player_id`."""
        return

    def notify_round_end(self, results: list[tuple[int, int]]) -> None:
        """Notify that the round has ended.

        Can be used for teardown code.
        """
        return



from itertools import chain, product
from unittest.mock import MagicMock, Mock

import pytest

from whist_backend.bet import Bet
from whist_backend.cards import Card, Rank, Suit
from whist_backend.player_strategy_protocol import PlayerStrategy
from whist_backend.round_simulator import (
    _Player,
    cyclic_shift,
    deal_hands,
    determine_trump_suit_and_hand_sizes,
    find_trick_winner,
)


class TestBet:
    @pytest.mark.parametrize(
        ["bet", "actual_result", "expected"],
        [
            (Bet(0, False), 1, -1),
            (Bet(3, False), 0, -3),
            (Bet(0, False), 0, 10),
            (Bet(2, False), 2, 12),
            (Bet(0, True), 1, -8),
            (Bet(0, True), 3, -24),
            (Bet(4, True), 2, -16),
            (Bet(0, True), 0, 20),
            (Bet(1, True), 1, 22),
            (Bet(3, True), 3, 26),
        ],
    )
    def test_bet_compute_score(
        self, bet: Bet, actual_result: int, expected: int
    ) -> None:
        assert bet.compute_score(actual_result) == expected

    @pytest.mark.parametrize(
        ["bet", "expected"],
        [
            (Bet(0, False), "0"),
            (Bet(1, False), "1"),
            (Bet(3, False), "3"),
            (Bet(10, False), "10"),
            (Bet(0, True), "CONFIDENT 0"),
            (Bet(1, True), "CONFIDENT 1"),
            (Bet(3, True), "CONFIDENT 3"),
            (Bet(10, True), "CONFIDENT 10"),
        ],
    )
    def test_bet_string(self, bet: Bet, expected: str) -> None:
        assert str(bet) == expected


class TestPlayer:
    def test_make_move_and_str(self) -> None:
        hand = [
            Card(Rank.ACE, Suit.SPADES),
            Card(Rank.TEN, Suit.SPADES),
            Card(Rank.TWO, Suit.CLUBS),
            Card(Rank.FOUR, Suit.DIAMONDS),
        ]
        strategy: PlayerStrategy = Mock()
        strategy.make_move.side_effect = [
            Card(Rank.ACE, Suit.SPADES),
            Card(Rank.TEN, Suit.SPADES),
        ]
        strategy.name = "TestBot"
        player = _Player(strategy, hand, Card(Rank.ACE, Suit.DIAMONDS), 1, 1)
        strategy.notify_round_start.assert_called_with(
            [
                Card(Rank.ACE, Suit.SPADES),
                Card(Rank.TEN, Suit.SPADES),
                Card(Rank.TWO, Suit.CLUBS),
                Card(Rank.FOUR, Suit.DIAMONDS),
            ],
            Suit.DIAMONDS,
            1,
            Card(Rank.ACE, Suit.DIAMONDS),
            1,
        )
        player.make_move(Suit.SPADES)
        strategy.make_move.assert_called_with(
            [Card(Rank.ACE, Suit.SPADES), Card(Rank.TEN, Suit.SPADES)]
        )
        player.make_move(Suit.SPADES)
        strategy.make_move.assert_called_with([Card(Rank.TEN, Suit.SPADES)])
        assert str(player) == "TestBot: hand  2♣, 4♦ (played  A♠,10♠)"

    @pytest.mark.parametrize(
        ["card"],
        [
            (Card(Rank.ACE, Suit.SPADES),),
            (Card(Rank.TEN, Suit.SPADES),),
            (Card(Rank.TWO, Suit.CLUBS),),
            (Card(Rank.FOUR, Suit.DIAMONDS),),
        ],
    )
    def test_valid_move_no_initial_suit(self, card: Card) -> None:
        hand = [
            Card(Rank.ACE, Suit.SPADES),
            Card(Rank.TEN, Suit.SPADES),
            Card(Rank.TWO, Suit.CLUBS),
            Card(Rank.FOUR, Suit.DIAMONDS),
        ]
        strategy: PlayerStrategy = Mock()
        strategy.make_move.side_effect = [card]
        player = _Player(strategy, hand, Mock(), Mock(), Mock())
        player.make_move(None)

    @pytest.mark.parametrize(
        ["card", "initial_suit"],
        [
            (Card(Rank.ACE, Suit.SPADES), Suit.SPADES),
            (Card(Rank.TEN, Suit.SPADES), Suit.HEARTS),
            (Card(Rank.TWO, Suit.CLUBS), Suit.CLUBS),
            (Card(Rank.FOUR, Suit.DIAMONDS), Suit.DIAMONDS),
            (Card(Rank.FOUR, Suit.DIAMONDS), Suit.HEARTS),
        ],
    )
    def test_valid_move_with_initial_suit(self, card: Card, initial_suit: Suit) -> None:
        hand = [
            Card(Rank.ACE, Suit.SPADES),
            Card(Rank.TEN, Suit.SPADES),
            Card(Rank.TWO, Suit.CLUBS),
            Card(Rank.FOUR, Suit.DIAMONDS),
        ]
        strategy: PlayerStrategy = Mock()
        strategy.make_move.side_effect = [card]
        player = _Player(strategy, hand, Mock(), Mock(), Mock())
        player.make_move(initial_suit)

    def test_invalid_move_no_initial_suit(self) -> None:
        hand = [
            Card(Rank.ACE, Suit.SPADES),
            Card(Rank.TEN, Suit.SPADES),
            Card(Rank.TWO, Suit.CLUBS),
            Card(Rank.FOUR, Suit.DIAMONDS),
        ]
        strategy: PlayerStrategy = Mock()
        strategy.make_move.side_effect = [Card(Rank.ACE, Suit.CLUBS)]
        player = _Player(strategy, hand, Mock(), Mock(), Mock())
        with pytest.raises(AssertionError):
            player.make_move(None)

    def test_invalid_move_with_initial_suit(self) -> None:
        hand = [
            Card(Rank.ACE, Suit.SPADES),
            Card(Rank.TEN, Suit.SPADES),
            Card(Rank.TWO, Suit.CLUBS),
            Card(Rank.FOUR, Suit.DIAMONDS),
        ]
        strategy: PlayerStrategy = Mock()
        strategy.make_move.side_effect = [Card(Rank.ACE, Suit.SPADES)]
        player = _Player(strategy, hand, Mock(), Mock(), Mock())
        with pytest.raises(AssertionError):
            player.make_move(Suit.DIAMONDS)

    def test_str(self) -> None:
        hand = [
            Card(Rank.ACE, Suit.SPADES),
            Card(Rank.TEN, Suit.SPADES),
            Card(Rank.TWO, Suit.CLUBS),
            Card(Rank.FOUR, Suit.DIAMONDS),
        ]
        strategy: PlayerStrategy = Mock()
        strategy.name = "TestBot"
        player = _Player(strategy, hand, Card(Rank.ACE, Suit.DIAMONDS), Mock(), 1)
        assert str(player) == "TestBot: hand  A♠,10♠, 2♣, 4♦"

    def test_make_valid_bet_no_disallowed(self) -> None:
        strategy: PlayerStrategy = Mock()
        strategy.name = "TestBot"
        strategy.make_bet.side_effect = [Bet(0, False)]
        player = _Player(strategy, Mock(), Mock(), Mock(), Mock())
        player.make_bet(None)

    def test_make_valid_bet(self) -> None:
        strategy: PlayerStrategy = Mock()
        strategy.name = "TestBot"
        strategy.make_bet.side_effect = [Bet(0, False)]
        player = _Player(strategy, Mock(), Mock(), Mock(), Mock())
        player.make_bet(1)

    def test_make_invalid_bet(self) -> None:
        strategy: PlayerStrategy = Mock()
        strategy.name = "TestBot"
        strategy.make_bet.side_effect = [Bet(0, False)]
        player = _Player(strategy, Mock(), Mock(), Mock(), Mock())
        with pytest.raises(AssertionError):
            player.make_bet(0)

    @pytest.mark.parametrize(
        ["initial_suit", "expected"],
        [
            (Suit.SPADES, [Card(Rank.ACE, Suit.SPADES), Card(Rank.TEN, Suit.SPADES)]),
            (
                Suit.HEARTS,
                [
                    Card(Rank.ACE, Suit.SPADES),
                    Card(Rank.TEN, Suit.SPADES),
                    Card(Rank.TWO, Suit.CLUBS),
                    Card(Rank.FOUR, Suit.DIAMONDS),
                ],
            ),
            (Suit.DIAMONDS, [Card(Rank.FOUR, Suit.DIAMONDS)]),
            (Suit.CLUBS, [Card(Rank.TWO, Suit.CLUBS)]),
            (
                None,
                [
                    Card(Rank.ACE, Suit.SPADES),
                    Card(Rank.TEN, Suit.SPADES),
                    Card(Rank.TWO, Suit.CLUBS),
                    Card(Rank.FOUR, Suit.DIAMONDS),
                ],
            ),
        ],
    )
    def test_get_allowed_cards(self, initial_suit: Suit, expected: list[Card]) -> None:
        hand = [
            Card(Rank.ACE, Suit.SPADES),
            Card(Rank.TEN, Suit.SPADES),
            Card(Rank.TWO, Suit.CLUBS),
            Card(Rank.FOUR, Suit.DIAMONDS),
        ]
        player = _Player(Mock(), hand, Mock(), Mock(), Mock())
        assert player.get_allowed_cards(initial_suit) == expected

    def test_notify_trick_won(self) -> None:
        strategy: PlayerStrategy = Mock()
        player = _Player(strategy, Mock(), Mock(), 1, 1)
        player.tricks_won = MagicMock()
        player.notify_trick_won(1)
        strategy.notify_trick_won.called_once_with(1)
        player.tricks_won.__iadd__.called_once_with(1)

    def test_notify_trick_not_won(self) -> None:
        strategy: PlayerStrategy = Mock()
        player = _Player(strategy, Mock(), Mock(), 1, 1)
        player.tricks_won = MagicMock()
        player.notify_trick_won(2)
        strategy.notify_trick_won.called_once_with(2)
        player.tricks_won.__iadd__.assert_not_called()


def test_deal_hands() -> None:
    deck = [Card(rank, suit) for rank, suit in product(Rank, Suit)]
    hands = deal_hands(4, 13, deck)
    assert all(len(hand) == 13 for hand in hands)
    assert len(set(chain(*hands))) == 52


@pytest.mark.parametrize(
    ["revealed_card", "expected"],
    [
        (Card(Rank.ACE, Suit.SPADES), (Suit.SPADES, 9)),
        (Card(Rank.TEN, Suit.DIAMONDS), (Suit.DIAMONDS, 10)),
        (Card(Rank.JACK, Suit.DIAMONDS), (Suit.DIAMONDS, 10)),
        (Card(Rank.FOUR, Suit.CLUBS), (Suit.CLUBS, 6)),
        (Card(Rank.KING, Suit.HEARTS), (Suit.HEARTS, 8)),
    ],
)
def test_determine_trump_suit_and_hand_sizes(
    revealed_card: Card, expected: tuple[Suit, int]
) -> None:
    assert determine_trump_suit_and_hand_sizes(revealed_card) == expected


@pytest.mark.parametrize(
    ["input_list", "amount", "expected"],
    [
        ([1, 2, 3, 4, 5], 0, [1, 2, 3, 4, 5]),
        ([1, 2, 3, 4, 5], 1, [2, 3, 4, 5, 1]),
        ([1, 2, 3, 4, 5], 2, [3, 4, 5, 1, 2]),
        ([1, 2, 3, 4, 5], 3, [4, 5, 1, 2, 3]),
        ([1, 2, 3, 4, 5], 4, [5, 1, 2, 3, 4]),
        ([1, 2, 3, 4, 5], 5, [1, 2, 3, 4, 5]),
    ],
)
def test_cyclic_shift(input_list: list[int], amount: int, expected: list[int]) -> None:
    assert cyclic_shift(input_list, amount) == expected


CARDS_AND_PLAYERS = [
    (0, Card(Rank.ACE, Suit.SPADES)),
    (1, Card(Rank.KING, Suit.SPADES)),
    (2, Card(Rank.TEN, Suit.DIAMONDS)),
    (3, Card(Rank.ACE, Suit.DIAMONDS)),
    (4, Card(Rank.TWO, Suit.HEARTS)),
]


@pytest.mark.parametrize(
    ["trump_suit", "initial_suit", "expected"],
    [
        (Suit.SPADES, Suit.SPADES, 0),
        (Suit.DIAMONDS, Suit.HEARTS, 3),
        (Suit.HEARTS, Suit.SPADES, 4),
        (Suit.HEARTS, Suit.HEARTS, 4),
        (Suit.CLUBS, Suit.HEARTS, 4),
        (Suit.CLUBS, Suit.SPADES, 0),
    ],
)
def test_find_trick_winner(trump_suit: Suit, initial_suit: Suit, expected: int) -> None:
    assert (
        find_trick_winner(CARDS_AND_PLAYERS, trump_suit, initial_suit) == expected
    )  # type: ignore



from .bet import Bet
from .cards import Card, Rank, Suit, new_deck, new_shuffled_deck
from .event_listener_protocol import WhistEventListener
from .player_strategy_protocol import PlayerStrategy
from .round_simulator import RoundSimulator as RoundSimulator

__version__ = "0.2.3"

__all__ = [
    "__version__",
    "Bet",
    "Card",
    "Rank",
    "Suit",
    "new_deck",
    "new_shuffled_deck",
    "WhistEventListener",
    "PlayerStrategy",
    "RoundSimulator",
]



from dataclasses import dataclass


@dataclass(frozen=True)
class Bet:
    """Each player must make a `Bet` at the beginning of each round."""

    number_of_tricks: int
    confident: bool

    def compute_score(self, actual_result: int) -> int:
        """Return the number of points the player should based on the actual outcome."""
        if actual_result == self.number_of_tricks:
            points = actual_result + 10
            return 2 * points if self.confident else points
        else:
            difference = abs(actual_result - self.number_of_tricks)
            return -8 * difference if self.confident else -1 * difference

    def __str__(self) -> str:
        """Return a string representation, eg `2` or `CONFIDENT 1`"""
        return ("CONFIDENT " if self.confident else "") + str(self.number_of_tricks)
